/*****************************************************************************

NAME:
   autodaemon .c -- self-starting daemon

ENTRY POINTS:
   client()	-- wrapper for the client end of a client/daemon pair
   server()	-- wrapper for the daemon end of a client/daemon pair
   daemonize()	-- make the current process a daemon

THEORY
   Consider the following problem: your program is called frequently to
answer some query.  Initializing the database to answer the question is
expensive, but answering questions is cheap.  You don't want to pay the
startup overhead on each call.  
   Solution: autodaemon.  The first time this code is called, it forks
a lighweight TCP/IP server as a daemon.  The daemon does the expensive
startup *one*.  All subsequent calls to the service query the daemon.
   With suitable pains taken, the process-hacking stuff in here could
be made portable to decade-old Unix boxes; I've done it before.  But
fuck that.  It's all about POSIX interfaces and Linux now, baby,

AUTHOR:
   Eric S. Raymond <esr@thyrsus.com>

******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include "autodaemon.h"
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/stat.h>	/* get umask(2) prototyped */

#if !defined(HAVE_SETSID) && defined(SIGTSTP)
#if defined(HAVE_TERMIOS_H)
#  include <termios.h>		/* for TIOCNOTTY under Linux */
#endif
#endif /* !defined(HAVE_SETSID) && defined(SIGTSTP) */

#define	SELF_ADDRESS	"127.0.0.1" 

int daemonize(void)
/* detach from control TTY, become process group leader, catch SIGCHLD */
{
    int fd;
    pid_t childpid;

    // if we are started by init (process 1) via /etc/inittab,
    // we needn't bother to detach from our process group context
    if (getppid() != 1) 
    {
	struct sigaction sa_new;

	/* Ignore BSD terminal stop signals */
	memset (&sa_new, 0, sizeof sa_new);
	sigemptyset (&sa_new.sa_mask);
	sa_new.sa_handler = SIG_IGN;
#ifdef 	SIGTTOU
	sigaction (SIGTTOU, &sa_new, NULL);
#endif
#ifdef	SIGTTIN
	sigaction (SIGTTIN, &sa_new, NULL);
#endif
#ifdef	SIGTSTP
	sigaction (SIGTSTP, &sa_new, NULL);
#endif

	// In case we were not started in the background, fork and let
	// the parent exit.  Guarantees that the child is not a process
	// group leader
	if ((childpid = fork()) < 0) {
	    return(AD_NOFORK);
	}
	else if (childpid > 0) 
	    exit(0);  /* parent */
  
	// Make ourselves the leader of a new process group with no
	// controlling terminal.
	if (setsid() < 0) {
	    return(AD_NOSETSID);
	}
  
	// lose controlling tty
	sigaction (SIGHUP, &sa_new, NULL);
	if ((childpid = fork()) < 0) {
	    return(AD_NOFORK);
	}
	else if (childpid > 0) {
	    exit(0); 	/* parent */
	}

    }

    // close any/all open file descriptors
#ifndef NOFILE
#define NOFILE	19	// educcated guess
#endif
    for (fd = NOFILE-1;  fd >= 0;  fd--)
	close(fd);	// not checking this should be safe, no writes

    // OK for these to fail, we're probably not going to use them anyway
    fd = open("/dev/null", O_RDONLY);	// stdin to /dev/null
    dup(fd);				// stdout to /dev/null
    dup(fd);				// stderr to /dev/null

    // move to root directory, so we don't prevent filesystem unmounts
    chdir("/");

    // set our umask to something reasonable (we hope) */
    umask(022);

    return(AD_SUCCESS);
}

// The rest of the world.

static int readline(int fd, char *ptr, int maxlen)
// equivalent of fgets(3) when fd is a stream socket
{
    int n, rc;
    char c;

    for (n = 1; n < maxlen; n++) {
	if ( (rc = read(fd, &c, 1)) == 1) {
	    *ptr++ = c;
	    if (c == '\n')
		break;
	} else if (rc == 0) {
	    if (n == 1)
		return(0);	// EOF, no data read
	    else
		break;		// EOF, some data was read
	} else
	    return(-1);
    }
  
    *ptr = 0;
    return(n);
}

static int writen(int fd, char *ptr, int nbytes) 
// substitute for write(2) when fd is a stream socket */
{
  int nleft, nwritten;
  
  nleft = nbytes;
  while (nleft > 0) {
    nwritten = write(fd, ptr, nleft);
    if (nwritten <= 0)
	return(nwritten);		//
    
    nleft -= nwritten;
    ptr   += nwritten;
  }
  return(nbytes - nleft);
}

// These just hide the socket from the caller.
// All this would be rather more elgant in an OO language.

static int client_socket;

static int client_read(char *buf, int maxlen)
// Read-from-server function to be handed to the client hook
{
    return readline(client_socket, buf, maxlen);
}

static int client_write(char *buf, int len)
// Write-to-server function to be handed to the client hook
{
    return writen(client_socket, buf, len);
}

static int server_socket;

static int server_read(char *buf, int maxlen)
// Read-from-client function to be handed to the server hook
{
    return readline(server_socket, buf, maxlen);
}

static int server_write(char *buf, int len)
// Write-to-client function to be handed to the server hook
{
    return writen(server_socket, buf, len);
}

//
// Entry points begin here 
//

int client(char *addr, int port,
	   int (*hook)(int (*read)(char *, int), int (*write)(char *, int), void *),
	   void *passin)
{
    struct sockaddr_in serv_addr;
    int status;

    // Fill in the structure "serv_addr" with the address of the
    // server that we want to connect with.
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = inet_addr(addr);
    serv_addr.sin_port        = htons(port);
  
    // Open a TCP socket (an Internet stream socket).
    if ( (client_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return(AD_NOSOCKET);	// errno will be set
    }
  
    // Connect to the server.
    if (connect(client_socket, (struct sockaddr *) &serv_addr,
		sizeof(serv_addr)) < 0) {
	return(AD_NOCONNECT);	// errno will be set
    }
  
    // run interpretation
    status = hook(client_read, client_write, passin);
  
    close(client_socket);
    return(status);
}

int server(int port,
	   int (*hook)(int (*read)(char *, int), int (*write)(char *, int), void *),
	   void *passin)
{
    int sockfd, clilen, childpid;
    struct sockaddr_in cli_addr, serv_addr;

     // Open a TCP socket (an Internet stream socket).
    if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	return(AD_NOSOCKET);	// errno will be set
    }
  
    // Bind our local address so that the client can send to us.
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family      = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port        = htons(port);
  
    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) {
	return(AD_NOBIND);	// errno will be set
    }
  
    if (listen(sockfd, 5) < 0) {
	return(AD_NOLISTEN);	// errno will be set
    }
                
    for ( ; ; ) {
	// Wait for a connection from a client process.
	// This is an example of a concurrent server.
	clilen = sizeof(cli_addr);
	server_socket = accept(sockfd, (struct sockaddr *) &cli_addr,
			   &clilen);
	if (server_socket < 0) {
	    return(AD_NOACCEPT);	// errno will be set
	}
    
	if ( (childpid = fork()) < 0) {
	    return(AD_NOFORK);		// errno will be set
	} else if (childpid == 0) {	// child process
	    close(sockfd);		// close original socket

	    // run interpreatation
	    _exit(hook(server_read, server_write, passin));

	    _exit(0);
	}

	close(server_socket);		// parent process
    }
    return(AD_SUCCESS);
}

int autodaemon(int port,
	       int (*clie)(int (*read)(char *,int), int (*write)(char *,int), void *),
	       int (*serv)(int (*read)(char *,int), int (*write)(char *,int), void *),
	       void *passin)
// Talk to a daemon version of ourself.  Fork one if it doesn't exist
{
    int status;

    // Try to connect to an existing server
    status = client(SELF_ADDRESS, port, clie, (void *)NULL);
    if (status == AD_SUCCESS)
	return(status);

    // No existing server?  Maybe we should spawn one.
    if (status == AD_NOCONNECT)
    {
	int status = server(port, serv, (void *)NULL);
	if (status == 0)
	    return client(SELF_ADDRESS, port, clie, (void *)NULL);
    }

    // Can't do anything right
    return(status);
}

#ifdef MAIN
#include <ctype.h>

#define MAXLINE	2048

#define	SERV_TCP_PORT	6060

/* host addr for server */

int server_guts(int (*read)(char *, int), int (*write)(char *, int), void *passin)
// sample server-side interpreter -- all-caps the inputs.
{
    int n;
    char line[MAXLINE], *cp;

    fprintf(stderr, "bogofilter server: started up\n");
    for ( ; ; ) {
	n = read(line, MAXLINE);
	if (n == 0)
	    return;		/* connection terminated */
	else if (n < 0) 
	    fprintf(stderr, "server_guts: readline error");

	fprintf(stderr, "bogofilter server: client sent %s", line);
	for (cp = line; *cp; cp++)
	    *cp = toupper(*cp);
    
	if (write(line, n) != n)
	    fprintf(stderr, "server_guts: writen error");
    }
    fprintf(stderr, "bogofilter server: signing off\n");
}

int client_guts(int (*read)(char *, int), int (*write)(char *, int), void *passin)
// sample client-side interpreter -- ship an input line, get back a line.
{
    int n;
    char sendline[MAXLINE], recvline[MAXLINE + 1];

    fprintf(stderr, "bogofilter client: started up\n");
    while (fgets(sendline, MAXLINE, stdin) != NULL) 
    {
	// Ship the line down
	n = strlen(sendline);
	if (write(sendline, n) != n) 
	    fprintf(stderr, "client_guts: write error on socket\n");
      
	// Now read a line from the socket and report it.
	n = read(recvline, MAXLINE);
	fprintf(stderr, "bogofilter client: server sent: %s", recvline);
	if (n < 0)
	    fprintf(stderr, "client_guts: readline error\n");
    }
  
    if (ferror(stdin))
	fprintf(stderr, "bogofilter client: error reading stdin\n");
    fprintf(stderr, "bogofilter client: signing off\n");
    return(AD_SUCCESS);
}

int main(int argc, char **argv)
{
    int status;
    if (argc > 1)
	status = server(SERV_TCP_PORT, server_guts, (void *)NULL);
    else
	status = client(SELF_ADDRESS, SERV_TCP_PORT, client_guts, (void *)NULL);
    if (status)
	fprintf(stderr, "Error %d, %s\n", status, strerror(errno)); 
    exit(0);
}

#endif /* MAIN */
